@@ -69,7 +69,7 @@ static char *substr(const char *head, const char *tail)
 	return buf;
 }
 
-static void parse_user(const char *t, char **name, char **email, unsigned long *date)
+static void parse_user(const char *t, char **name, char **email, unsigned long *date, int *tz)
 {
 	struct ident_split ident;
 	unsigned email_len;
@@ -83,6 +83,8 @@ static void parse_user(const char *t, char **name, char **email, unsigned long *
 
 		if (ident.date_begin)
 			*date = strtoul(ident.date_begin, NULL, 10);
+		if (ident.tz_begin)
+			*tz = atoi(ident.tz_begin);
 	}
 }
 
@@ -139,7 +141,7 @@ struct commitinfo *cgit_parse_commit(struct commit *commit)
 		return ret;
 
 	if (!skip_prefix(p, "tree ", &p))
-		die("Bad commit: %s", sha1_to_hex(commit->object.sha1));
+		die("Bad commit: %s", oid_to_hex(&commit->object.oid));
 	p += sha1hex_len + 1;
 
 	while (skip_prefix(p, "parent ", &p))
@@ -147,13 +149,13 @@ struct commitinfo *cgit_parse_commit(struct commit *commit)
 
 	if (p && skip_prefix(p, "author ", &p)) {
 		parse_user(p, &ret->author, &ret->author_email,
-			&ret->author_date);
+			&ret->author_date, &ret->author_tz);
 		p = next_header_line(p);
 	}
 
 	if (p && skip_prefix(p, "committer ", &p)) {
 		parse_user(p, &ret->committer, &ret->committer_email,
-			&ret->committer_date);
+			&ret->committer_date, &ret->committer_tz);
 		p = next_header_line(p);
 	}
 
@@ -199,7 +201,7 @@ struct taginfo *cgit_parse_tag(struct tag *tag)
 	const char *p;
 	struct taginfo *ret = NULL;
 
-	data = read_sha1_file(tag->object.sha1, &type, &size);
+	data = read_sha1_file(tag->object.oid.hash, &type, &size);
 	if (!data || type != OBJ_TAG)
 		goto cleanup;
 
@@ -208,7 +210,7 @@ struct taginfo *cgit_parse_tag(struct tag *tag)
 	for (p = data; !end_of_header(p); p = next_header_line(p)) {
 		if (skip_prefix(p, "tagger ", &p)) {
 			parse_user(p, &ret->tagger, &ret->tagger_email,
-				&ret->tagger_date);
+				&ret->tagger_date, &ret->tagger_tz);
 		}
 	}
 
