@@ -19,28 +19,31 @@ void cgit_print_commit(char *hex, const char *prefix)
 	struct commitinfo *info, *parent_info;
 	struct commit_list *p;
 	struct strbuf notes = STRBUF_INIT;
-	unsigned char sha1[20];
+	struct object_id oid;
 	char *tmp, *tmp2;
 	int parents = 0;
 
 	if (!hex)
 		hex = ctx.qry.head;
 
-	if (get_sha1(hex, sha1)) {
-		cgit_print_error("Bad object id: %s", hex);
+	if (get_oid(hex, &oid)) {
+		cgit_print_error_page(400, "Bad request",
+				"Bad object id: %s", hex);
 		return;
 	}
-	commit = lookup_commit_reference(sha1);
+	commit = lookup_commit_reference(oid.hash);
 	if (!commit) {
-		cgit_print_error("Bad commit reference: %s", hex);
+		cgit_print_error_page(404, "Not found",
+				"Bad commit reference: %s", hex);
 		return;
 	}
 	info = cgit_parse_commit(commit);
 
-	format_display_notes(sha1, &notes, PAGE_ENCODING, 0);
+	format_display_notes(oid.hash, &notes, PAGE_ENCODING, 0);
 
 	load_ref_decorations(DECORATE_FULL_REFS);
 
+	cgit_print_layout_start();
 	cgit_print_diff_ctrls();
 	html("<table summary='commit info' class='commit-info'>\n");
 	html("<tr><th>author</th><td>");
@@ -52,7 +55,8 @@ void cgit_print_commit(char *hex, const char *prefix)
 	}
 	cgit_close_filter(ctx.repo->email_filter);
 	html("</td><td class='right'>");
-	cgit_print_date(info->author_date, FMT_LONGDATE, ctx.cfg.local_time);
+	html_txt(show_date(info->author_date, info->author_tz,
+				cgit_date_mode(DATE_ISO8601)));
 	html("</td></tr>\n");
 	html("<tr><th>committer</th><td>");
 	cgit_open_filter(ctx.repo->email_filter, info->committer_email, "commit");
@@ -63,17 +67,18 @@ void cgit_print_commit(char *hex, const char *prefix)
 	}
 	cgit_close_filter(ctx.repo->email_filter);
 	html("</td><td class='right'>");
-	cgit_print_date(info->committer_date, FMT_LONGDATE, ctx.cfg.local_time);
+	html_txt(show_date(info->committer_date, info->committer_tz,
+				cgit_date_mode(DATE_ISO8601)));
 	html("</td></tr>\n");
 	html("<tr><th>commit</th><td colspan='2' class='sha1'>");
-	tmp = sha1_to_hex(commit->object.sha1);
+	tmp = oid_to_hex(&commit->object.oid);
 	cgit_commit_link(tmp, NULL, NULL, ctx.qry.head, tmp, prefix);
 	html(" (");
 	cgit_patch_link("patch", NULL, NULL, NULL, tmp, prefix);
 	html(")</td></tr>\n");
 	html("<tr><th>tree</th><td colspan='2' class='sha1'>");
 	tmp = xstrdup(hex);
-	cgit_tree_link(sha1_to_hex(commit->tree->object.sha1), NULL, NULL,
+	cgit_tree_link(oid_to_hex(&commit->tree->object.oid), NULL, NULL,
 		       ctx.qry.head, tmp, NULL);
 	if (prefix) {
 		html(" /");
@@ -82,7 +87,7 @@ void cgit_print_commit(char *hex, const char *prefix)
 	free(tmp);
 	html("</td></tr>\n");
 	for (p = commit->parents; p; p = p->next) {
-		parent = lookup_commit_reference(p->item->object.sha1);
+		parent = lookup_commit_reference(p->item->object.oid.hash);
 		if (!parent) {
 			html("<tr><td colspan='3'>");
 			cgit_print_error("Error reading parent commit");
@@ -91,7 +96,7 @@ void cgit_print_commit(char *hex, const char *prefix)
 		}
 		html("<tr><th>parent</th>"
 		     "<td colspan='2' class='sha1'>");
-		tmp = tmp2 = sha1_to_hex(p->item->object.sha1);
+		tmp = tmp2 = oid_to_hex(&p->item->object.oid);
 		if (ctx.repo->enable_subject_links) {
 			parent_info = cgit_parse_commit(parent);
 			tmp2 = parent_info->subject;
@@ -99,7 +104,7 @@ void cgit_print_commit(char *hex, const char *prefix)
 		cgit_commit_link(tmp2, NULL, NULL, ctx.qry.head, tmp, prefix);
 		html(" (");
 		cgit_diff_link("diff", NULL, NULL, ctx.qry.head, hex,
-			       sha1_to_hex(p->item->object.sha1), prefix);
+			       oid_to_hex(&p->item->object.oid), prefix);
 		html(")</td></tr>");
 		parents++;
 	}
@@ -132,11 +137,12 @@ void cgit_print_commit(char *hex, const char *prefix)
 	}
 	if (parents < 3) {
 		if (parents)
-			tmp = sha1_to_hex(commit->parents->item->object.sha1);
+			tmp = oid_to_hex(&commit->parents->item->object.oid);
 		else
 			tmp = NULL;
 		cgit_print_diff(ctx.qry.sha1, tmp, prefix, 0, 0);
 	}
 	strbuf_release(&notes);
 	cgit_free_commitinfo(info);
+	cgit_print_layout_end();
 }
