@@ -12,8 +12,8 @@
 #include "ui-shared.h"
 #include "ui-ssdiff.h"
 
-unsigned char old_rev_sha1[20];
-unsigned char new_rev_sha1[20];
+struct object_id old_rev_oid[1];
+struct object_id new_rev_oid[1];
 
 static int files, slots;
 static int total_adds, total_rems, max_changes;
@@ -21,8 +21,8 @@ static int lines_added, lines_removed;
 
 static struct fileinfo {
 	char status;
-	unsigned char old_sha1[20];
-	unsigned char new_sha1[20];
+	struct object_id old_oid[1];
+	struct object_id new_oid[1];
 	unsigned short old_mode;
 	unsigned short new_mode;
 	char *old_path;
@@ -36,6 +36,7 @@ static struct fileinfo {
 
 static int use_ssdiff = 0;
 static struct diff_filepair *current_filepair;
+static const char *current_prefix;
 
 struct diff_filespec *cgit_get_current_old_file(void)
 {
@@ -82,15 +83,15 @@ static void print_fileinfo(struct fileinfo *info)
 
 	html("<tr>");
 	htmlf("<td class='mode'>");
-	if (is_null_sha1(info->new_sha1)) {
+	if (is_null_oid(info->new_oid)) {
 		cgit_print_filemode(info->old_mode);
 	} else {
 		cgit_print_filemode(info->new_mode);
 	}
 
 	if (info->old_mode != info->new_mode &&
-	    !is_null_sha1(info->old_sha1) &&
-	    !is_null_sha1(info->new_sha1)) {
+	    !is_null_oid(info->old_oid) &&
+	    !is_null_oid(info->new_oid)) {
 		html("<span class='modechange'>[");
 		cgit_print_filemode(info->old_mode);
 		html("]</span>");
@@ -132,15 +133,34 @@ static void count_diff_lines(char *line, int len)
 	}
 }
 
+static int show_filepair(struct diff_filepair *pair)
+{
+	/* Always show if we have no limiting prefix. */
+	if (!current_prefix)
+		return 1;
+
+	/* Show if either path in the pair begins with the prefix. */
+	if (starts_with(pair->one->path, current_prefix) ||
+	    starts_with(pair->two->path, current_prefix))
+		return 1;
+
+	/* Otherwise we don't want to show this filepair. */
+	return 0;
+}
+
 static void inspect_filepair(struct diff_filepair *pair)
 {
 	int binary = 0;
 	unsigned long old_size = 0;
 	unsigned long new_size = 0;
+
+	if (!show_filepair(pair))
+		return;
+
 	files++;
 	lines_added = 0;
 	lines_removed = 0;
-	cgit_diff_files(pair->one->sha1, pair->two->sha1, &old_size, &new_size,
+	cgit_diff_files(&pair->one->oid, &pair->two->oid, &old_size, &new_size,
 			&binary, 0, ctx.qry.ignorews, count_diff_lines);
 	if (files >= slots) {
 		if (slots == 0)
@@ -150,8 +170,8 @@ static void inspect_filepair(struct diff_filepair *pair)
 		items = xrealloc(items, slots * sizeof(struct fileinfo));
 	}
 	items[files-1].status = pair->status;
-	hashcpy(items[files-1].old_sha1, pair->one->sha1);
-	hashcpy(items[files-1].new_sha1, pair->two->sha1);
+	oidcpy(items[files-1].old_oid, &pair->one->oid);
+	oidcpy(items[files-1].new_oid, &pair->two->oid);
 	items[files-1].old_mode = pair->one->mode;
 	items[files-1].new_mode = pair->two->mode;
 	items[files-1].old_path = xstrdup(pair->one->path);
@@ -167,8 +187,8 @@ static void inspect_filepair(struct diff_filepair *pair)
 	total_rems += lines_removed;
 }
 
-static void cgit_print_diffstat(const unsigned char *old_sha1,
-				const unsigned char *new_sha1,
+static void cgit_print_diffstat(const struct object_id *old_oid,
+				const struct object_id *new_oid,
 				const char *prefix)
 {
 	int i;
@@ -184,7 +204,7 @@ static void cgit_print_diffstat(const unsigned char *old_sha1,
 	html("</div>");
 	html("<table summary='diffstat' class='diffstat'>");
 	max_changes = 0;
-	cgit_diff_tree(old_sha1, new_sha1, inspect_filepair, prefix,
+	cgit_diff_tree(old_oid, new_oid, inspect_filepair, prefix,
 		       ctx.qry.ignorews);
 	for (i = 0; i<files; i++)
 		print_fileinfo(&items[i]);
@@ -218,8 +238,8 @@ static void print_line(char *line, int len)
 	line[len-1] = c;
 }
 
-static void header(unsigned char *sha1, char *path1, int mode1,
-		   unsigned char *sha2, char *path2, int mode2)
+static void header(const struct object_id *oid1, char *path1, int mode1,
+		   const struct object_id *oid2, char *path2, int mode2)
 {
 	char *abbrev1, *abbrev2;
 	int subproject;
@@ -238,8 +258,8 @@ static void header(unsigned char *sha1, char *path1, int mode1,
 		htmlf("<br/>deleted file mode %.6o", mode1);
 
 	if (!subproject) {
-		abbrev1 = xstrdup(find_unique_abbrev(sha1, DEFAULT_ABBREV));
-		abbrev2 = xstrdup(find_unique_abbrev(sha2, DEFAULT_ABBREV));
+		abbrev1 = xstrdup(find_unique_abbrev(oid1->hash, DEFAULT_ABBREV));
+		abbrev2 = xstrdup(find_unique_abbrev(oid2->hash, DEFAULT_ABBREV));
 		htmlf("<br/>index %s..%s", abbrev1, abbrev2);
 		free(abbrev1);
 		free(abbrev2);
@@ -248,24 +268,24 @@ static void header(unsigned char *sha1, char *path1, int mode1,
 			if (mode2 != mode1)
 				htmlf("..%.6o", mode2);
 		}
-		if (is_null_sha1(sha1)) {
+		if (is_null_oid(oid1)) {
 			path1 = "dev/null";
 			html("<br/>--- /");
 		} else
 			html("<br/>--- a/");
 		if (mode1 != 0)
 			cgit_tree_link(path1, NULL, NULL, ctx.qry.head,
-				       sha1_to_hex(old_rev_sha1), path1);
+				       oid_to_hex(old_rev_oid), path1);
 		else
 			html_txt(path1);
-		if (is_null_sha1(sha2)) {
+		if (is_null_oid(oid2)) {
 			path2 = "dev/null";
 			html("<br/>+++ /");
 		} else
 			html("<br/>+++ b/");
 		if (mode2 != 0)
 			cgit_tree_link(path2, NULL, NULL, ctx.qry.head,
-				       sha1_to_hex(new_rev_sha1), path2);
+				       oid_to_hex(new_rev_oid), path2);
 		else
 			html_txt(path2);
 	}
@@ -279,25 +299,28 @@ static void filepair_cb(struct diff_filepair *pair)
 	int binary = 0;
 	linediff_fn print_line_fn = print_line;
 
+	if (!show_filepair(pair))
+		return;
+
 	current_filepair = pair;
 	if (use_ssdiff) {
 		cgit_ssdiff_header_begin();
 		print_line_fn = cgit_ssdiff_line_cb;
 	}
-	header(pair->one->sha1, pair->one->path, pair->one->mode,
-	       pair->two->sha1, pair->two->path, pair->two->mode);
+	header(&pair->one->oid, pair->one->path, pair->one->mode,
+	       &pair->two->oid, pair->two->path, pair->two->mode);
 	if (use_ssdiff)
 		cgit_ssdiff_header_end();
 	if (S_ISGITLINK(pair->one->mode) || S_ISGITLINK(pair->two->mode)) {
 		if (S_ISGITLINK(pair->one->mode))
-			print_line_fn(fmt("-Subproject %s", sha1_to_hex(pair->one->sha1)), 52);
+			print_line_fn(fmt("-Subproject %s", oid_to_hex(&pair->one->oid)), 52);
 		if (S_ISGITLINK(pair->two->mode))
-			print_line_fn(fmt("+Subproject %s", sha1_to_hex(pair->two->sha1)), 52);
+			print_line_fn(fmt("+Subproject %s", oid_to_hex(&pair->two->oid)), 52);
 		if (use_ssdiff)
 			cgit_ssdiff_footer();
 		return;
 	}
-	if (cgit_diff_files(pair->one->sha1, pair->two->sha1, &old_size,
+	if (cgit_diff_files(&pair->one->oid, &pair->two->oid, &old_size,
 			    &new_size, &binary, ctx.qry.context,
 			    ctx.qry.ignorews, print_line_fn))
 		cgit_print_error("Error running diff");
@@ -317,7 +340,7 @@ void cgit_print_diff_ctrls(void)
 
 	html("<div class='cgit-panel'>");
 	html("<b>diff options</b>");
-	html("<form method='get' action='.'>");
+	html("<form method='get'>");
 	cgit_add_hidden_formfields(1, 0, ctx.qry.page);
 	html("<table>");
 	html("<tr><td colspan='2'/></tr>");
@@ -365,37 +388,53 @@ void cgit_print_diff(const char *new_rev, const char *old_rev,
 	const unsigned char *old_tree_sha1, *new_tree_sha1;
 	diff_type difftype;
 
+	/*
+	 * If "follow" is set then the diff machinery needs to examine the
+	 * entire commit to detect renames so we must limit the paths in our
+	 * own callbacks and not pass the prefix to the diff machinery.
+	 */
+	if (ctx.qry.follow && ctx.cfg.enable_follow_links) {
+		current_prefix = prefix;
+		prefix = "";
+	} else {
+		current_prefix = NULL;
+	}
+
 	if (!new_rev)
 		new_rev = ctx.qry.head;
-	if (get_sha1(new_rev, new_rev_sha1)) {
-		cgit_print_error("Bad object name: %s", new_rev);
+	if (get_oid(new_rev, new_rev_oid)) {
+		cgit_print_error_page(404, "Not found",
+			"Bad object name: %s", new_rev);
 		return;
 	}
-	commit = lookup_commit_reference(new_rev_sha1);
+	commit = lookup_commit_reference(new_rev_oid->hash);
 	if (!commit || parse_commit(commit)) {
-		cgit_print_error("Bad commit: %s", sha1_to_hex(new_rev_sha1));
+		cgit_print_error_page(404, "Not found",
+			"Bad commit: %s", oid_to_hex(new_rev_oid));
 		return;
 	}
-	new_tree_sha1 = commit->tree->object.sha1;
+	new_tree_sha1 = commit->tree->object.oid.hash;
 
 	if (old_rev) {
-		if (get_sha1(old_rev, old_rev_sha1)) {
-			cgit_print_error("Bad object name: %s", old_rev);
+		if (get_oid(old_rev, old_rev_oid)) {
+			cgit_print_error_page(404, "Not found",
+				"Bad object name: %s", old_rev);
 			return;
 		}
 	} else if (commit->parents && commit->parents->item) {
-		hashcpy(old_rev_sha1, commit->parents->item->object.sha1);
+		oidcpy(old_rev_oid, &commit->parents->item->object.oid);
 	} else {
-		hashclr(old_rev_sha1);
+		oidclr(old_rev_oid);
 	}
 
-	if (!is_null_sha1(old_rev_sha1)) {
-		commit2 = lookup_commit_reference(old_rev_sha1);
+	if (!is_null_oid(old_rev_oid)) {
+		commit2 = lookup_commit_reference(old_rev_oid->hash);
 		if (!commit2 || parse_commit(commit2)) {
-			cgit_print_error("Bad commit: %s", sha1_to_hex(old_rev_sha1));
+			cgit_print_error_page(404, "Not found",
+				"Bad commit: %s", oid_to_hex(old_rev_oid));
 			return;
 		}
-		old_tree_sha1 = commit2->tree->object.sha1;
+		old_tree_sha1 = commit2->tree->object.oid.hash;
 	} else {
 		old_tree_sha1 = NULL;
 	}
@@ -425,8 +464,10 @@ void cgit_print_diff(const char *new_rev, const char *old_rev,
 	difftype = ctx.qry.has_difftype ? ctx.qry.difftype : ctx.cfg.difftype;
 	use_ssdiff = difftype == DIFF_SSDIFF;
 
-	if (show_ctrls)
+	if (show_ctrls) {
+		cgit_print_layout_start();
 		cgit_print_diff_ctrls();
+	}
 
 	/*
 	 * Clicking on a link to a file in the diff stat should show a diff
@@ -438,7 +479,7 @@ void cgit_print_diff(const char *new_rev, const char *old_rev,
 	if (difftype == DIFF_STATONLY)
 		ctx.qry.difftype = ctx.cfg.difftype;
 
-	cgit_print_diffstat(old_rev_sha1, new_rev_sha1, prefix);
+	cgit_print_diffstat(old_rev_oid, new_rev_oid, prefix);
 
 	if (difftype == DIFF_STATONLY)
 		return;
@@ -449,9 +490,12 @@ void cgit_print_diff(const char *new_rev, const char *old_rev,
 		html("<table summary='diff' class='diff'>");
 		html("<tr><td>");
 	}
-	cgit_diff_tree(old_rev_sha1, new_rev_sha1, filepair_cb, prefix,
+	cgit_diff_tree(old_rev_oid, new_rev_oid, filepair_cb, prefix,
 		       ctx.qry.ignorews);
 	if (!use_ssdiff)
 		html("</td></tr>");
 	html("</table>");
+
+	if (show_ctrls)
+		cgit_print_layout_end();
 }
